import asyncio
import inspect
import time
from datetime import datetime, timedelta
from threading import Thread
from typing import Callable, Any

import pyttman


class TimeTrigger:
    """
    TimeTrigger class

    The TimeTrigger class acts like a timed clock
    which can be wound up using different measurements
    and units of time. It can be set to trigger
    at a certain time, delay a certain amount of time
    has passed, be a recurring trigger with a defined
    interval of delay.

    """
    timeformat = "%H:%M:%S"
    day_int = {
        "day": (0, 1, 2, 3, 4, 5, 6),
        "monday": (0,),
        "tuesday": (1,),
        "wednesday": (2,),
        "thursday": (3,),
        "friday": (4,),
        "saturday": (5,),
        "sunday": (6,),
    }

    timeunits = {
        "hour": timedelta(hours=1),
        "minute": timedelta(minutes=1),
        "second": timedelta(seconds=1),
    }

    def __init__(self, at: str = None, every: str = None,
                 delay=None, exactly_at: datetime = None):
        """
        Configures the TimeTrigger object according
        to provided arguments.

        :param every: str, a day, every day or timeunit.
            Valid parameters:
                "day", "monday", "tuesday", "wednesday"
                "thursday", "friday", "saturday"
                "sunday", "hour", "minute", "second"

        :param at: str, time of scheduling.
                  Valid in HH:MM:SS format or HH:MM format.
                  Example: "10:00" or "10:00:22". Seconds (SS)
                  defaults to 00 if not specified.
        :param exactly_at: datetime object which will be set to
                           self.next_trigger if present, bypassing
                           the other parameters every, at, and delay.
                           Useful for scenarios when a datetime
                           object is created outside by the creator
                           for a specific day and time when the job
                           is to be run.
        """
        self.amount_of_runs = 0
        self.next_trigger: datetime = datetime.now()
        self.last_trigger: datetime = None
        self.reoccurring: bool = False
        self.timedelta_interval: timedelta = None
        self.days_to_run: tuple[int] = None
        self.delay = None

        if exactly_at:
            self.next_trigger = exactly_at
        if at:
            at = self.__validate_timestring(at)
            parsed_time = datetime.strptime(at, self.timeformat)
            self.next_trigger = self.next_trigger.replace(
                hour=parsed_time.hour,
                minute=parsed_time.minute,
                second=parsed_time.second)
        elif delay:
            delay = self.__validate_timestring(delay)
            parsed_time = datetime.strptime(delay, self.timeformat)
            self.delay = timedelta(hours=parsed_time.hour,
                                   minutes=parsed_time.minute,
                                   seconds=parsed_time.second)
        if every:
            self.reoccurring = True
            if not (_repeat_every := self.day_int.get(every)):
                if not (_repeat_every := self.timeunits.get(every)):
                    raise ValueError(
                        f"'{every}' is an invalid value for 'every'")
            if isinstance(_repeat_every, tuple):
                # It's a day of week, or every day in the week
                self.days_to_run = _repeat_every
            elif isinstance(_repeat_every, timedelta):
                # It's every minute, hour or second
                self.timedelta_interval = _repeat_every

        if not any((at, every, delay, exactly_at)):
            raise AttributeError("TimeTrigger needs at least one rule "
                                 "for scheduling: 'every', 'at', 'delay', "
                                 "'exactly_at'")

        self.reset()

    def __repr__(self):
        return f"TimeTrigger(" \
               f"next_trigger={self.next_trigger}, " \
               f"last_trigger={self.last_trigger}, " \
               f"amount_of_runs={self.amount_of_runs}, " \
               f"reccuring={self.reoccurring})" \


    @staticmethod
    def __validate_timestring(timestr: str) -> str:
        if len(timestr.split(":")) < 3:
            return f"{timestr}:00"
        return timestr

    def is_pulled(self):
        if datetime.now() >= self.next_trigger:
            if self.reoccurring:
                self.reset()
            self.amount_of_runs += 1
            self.last_trigger = datetime.now()
            return True
        return False

    def reset(self) -> None:
        """
        Reset itself to a future point in time,
        based on the rules provided at instantiation.

        If for example it is configured to run
        every day, the next day in line will
        be assigned to the datetime object in
        self.next_trigger.

        If a timedelta is the configured interval,
        the self.next_trigger will be mutated with
        this value.
        :returns: None
        """
        if self.days_to_run:
            if datetime.now() >= self.next_trigger:
                self.next_trigger += timedelta(days=1)
            while not self.next_trigger.weekday() in self.days_to_run:
                self.next_trigger += timedelta(days=1)
        elif self.timedelta_interval:
            self.next_trigger += self.timedelta_interval
        if self.delay:
            self.next_trigger += self.delay


class Job(Thread):
    """
    Class subclassing Thread, representing a scheduled
    function, class or other callable.

    The Job class is not designed to be accessed by
    users but is internally managed in the schedule
    API using Pyttman's scheduler.

    The Job class is runnable as a separate
    thread, thus leaving any function that it's
    calling non-blocking.

    The Job can run both async and sync functions
    both as the main callable and recipients.

    Users can provide which loop to schedule an async task
    in, but it is not mandatory - this may cause problems
    however in context managed async tasks which requires
    access to the loop to retrieve tasks themselves.

    TL;DR - if problems occur when 'async_loop' argument
    is omitted - provide the loop for where to schedule
    the recipient coroutine.
    """

    def __init__(self, func: Callable,
                 is_async: bool,
                 trigger: TimeTrigger,
                 recipient: Callable,
                 func_name: str,
                 async_loop: Any =None,
                 return_self: bool = False):
        super().__init__()
        self.kwargs = None
        self.func = func
        self.is_async = is_async
        self.func_name = func_name
        self.trigger = trigger
        self.recipient = recipient
        self.return_self = return_self
        self.time_to_die = False
        self.error = None
        self.result = None
        self.async_loop = async_loop
        self._running = False

    def __repr__(self):
        error = self.error if self.error else None
        name = type(self.error).__name__
        error_str = f"{name}({error})"

        return f"Job(" \
               f"func_name={self.func_name}, " \
               f"is_async={self.is_async}, " \
               f"running={self.running}, " \
               f"recipient={self.recipient.__name__}, " \
               f"native_id={self.native_id}, " \
               f"trigger={self.trigger}, " \
               f"result='{self.result}', " \
               f"error={error_str}"

    def run(self) -> None:
        """
        Thread overloaded method
        Await the TimeTrigger object to fire on
        it's point in time, and call the callable
        passed as self.func.
        """
        self._running = True

        while True:
            if self.time_to_die:
                pyttman.logger.log(
                    f"Job '{self.native_id}' got a "
                    f"graceful kill signal, shutting "
                    f"down.")
                break

            # Evaluate if self.func and / or recipient is async or not
            if self.trigger.is_pulled():
                try:
                    # Evaluate whether main callable and/or recipient is async
                    # Run it with async.run if no loop is defined, otherwise
                    # add it to existing loop as a created task
                    if self.is_async:
                        if not self.async_loop:
                            self.result = asyncio.run(self.func())
                        else:
                            self.async_loop.create_task(self.func())
                    else:
                        self.result = self.func()
                except Exception as e:
                    pyttman.logger.log(f"The schedule job '{self.name}' "
                                       f"raised {type(e).__name__}"
                                       f"('{str(e)}') upon executing it",
                                       level="error")
                    self.error = e
                    break

                # Call the recipient function with the job output
                output = self if self.return_self else self.result
                try:
                    if inspect.iscoroutinefunction(self.recipient):
                        if self.async_loop:
                            self.async_loop.create_task(self.recipient(output))
                        else:
                            asyncio.run(self.recipient(output))
                    else:
                        self.recipient(output)
                except Exception as e:
                    pyttman.logger.log(f"The schedule job '{self.name}' "
                                       f"ran OK but the recipient function "
                                       f"{self.recipient} raised "
                                       f"{type(e).__name__}"
                                       f"('{str(e)}') ", level="error")
                    break
            if not self.trigger.reoccurring and self.trigger.amount_of_runs > 0:
                break
            time.sleep(0.01)
        self._running = False

    @property
    def running(self) -> bool:
        return self._running

    def kill_gracefully(self) -> None:
        """
        Provices a method to set the self.time_to_die
        to True to signal it's time to check out
        """
        self.time_to_die = True



"""
This module defines the base class for the Storage API
in Pyttman.
"""
import abc
from abc import ABC
from collections import UserDict
from typing import Any


class AbstractStorage(UserDict, abc.ABC):
    """
    Abstract storage class.
    Do not subclass this class for custom
    storage classes - subclass the 'Storage'
    class instead.
    """

    @abc.abstractmethod
    def __repr__(self):
        pass

    @abc.abstractmethod
    def __contains__(self, item):
        pass

    @abc.abstractmethod
    def __getitem__(self, item):
        pass

    @abc.abstractmethod
    def put(self, key: Any, item: Any):
        pass

    @abc.abstractmethod
    def get(self, key) -> Any:
        pass

    @abc.abstractmethod
    def dump(self):
        pass

    @abc.abstractmethod
    def synchronize(self):
        pass


class BaseStorage(AbstractStorage, ABC):
    """
    Base Storage class for cache based storage in
    the Pyttman framework.

    The API implements the dictionary interface
    for ease of use, and also provides the 'put'
    method for semantic improvements.

    Subclasses can easily integrate a file or
    database storage backend by implementing
    the 'synchronize' method.
    """
    def __repr__(self):
        return f"{self.__class__.__name__}(keys={self.data.keys()})"

    def __contains__(self, item):
        return item in self.data

    def __getitem__(self, item):
        """
        Behaves just like dict.__getitem__.
        :param item: Any, key for requested object
        :return: Any
        :raises: KeyError if key not present in self.data
        """
        try:
            return self.data[item]
        except KeyError:
            raise KeyError("Pyttman.Storage: No item stored "
                           f"under key which matches '{item}'")

    def put(self, key: Any, item: Any):
        """
        Store an object in the Storage object, equivalent to
        `my_dict[key] = value"
        :param key: Any, key for the stored object
        :param item: Any, the actual object to store
        :return: None
        """
        self.data[key] = item

    def get(self, key) -> Any:
        """
        Get a stored object by provided key.
        This method does not raise an error
        like __getitem__ in dict, but behaves
        just like .get on dict objects.
        :param key: Any, key for requested object
        :return: Any
        """
        return self.data.get(key)


class Storage(BaseStorage):
    def dump(self):
        """
        Dump the storage to set backend type.
        Different backend types are available
        in storage.engines
        :return:
        """
        raise NotImplementedError("Dumping the Storage object is not "
                                  "implemented in this version of Pyttman, "
                                  "but will be available in a future release.")

    def synchronize(self):
        """
        Synchronize the storage object with set
        storage engine. In case of the SqlEngine,
        this would synchronize the Storage object
        with an SQL database.
        :return:
        """
        raise NotImplementedError("Synchronizing is not implemented "
                                  "in this version of Pyttman, but will "
                                  "be available in a future release.")



import pyttman
import pyttman.tools.pyttmancli.intents as intents
from pyttman.core.internals import PyttmanApp
from pyttman.core.ability import Ability


class PyttmanCli(Ability):
    """
    Encapsulates the Pyttman CLI tool 'pyttman'
    used in the terminal by framework users.
    """
    intents = (
        intents.CreateNewApp,
        intents.RunAppInDevMode,
        intents.RunAppInClientMode,
        intents.CreateNewAbilityIntent,
        intents.ShellMode,
        intents.VersionInfo,
        intents.RunFile,
    )

    description = f"\nPyttman v{pyttman.__version__}\n\n" \
                  "For help about a commend, type pyttman help [command]" \
                  f"\n\nSupported commands:\n"

    def before_create(self):
        self.storage.put("app", None)
        self.storage.put("ready", False)
        self.storage.put("template_url",
                         "https://github.com/Hashmap-Software-Agency/"
                         "pyttman-project-template/archive/refs/heads/main.zip")

    def run_application(self) -> None:
        """
        Runs a Pyttman application with its Runner context
        as provided.
        :return: None
        """
        # noinspection PyTypeChecker, PyUnusedLocal
        # #(used for attribute access in completion)
        app: PyttmanApp = None
        if (app := self.storage.get("app")) is not None:
            if not app.client.message_router.abilities:
                print("There are no abilities loaded, the app will not "
                      "respond to any messages. Create abilities by " 
                      "running 'pyttman new ability' in the terminal.\n"
                      "Next up, add them to your app's settings.py file "
                      "under the ABILITIES key.")
                exit(0)
            print(f"- [Pyttman]: Ability classes loaded: "
                  f"{app.client.message_router.abilities}")
            app.start()
        else:
            raise RuntimeError("No Runner provided, app cannot start. "
                               "Exiting...")




import functools
import traceback

import pyttman


class PyttmanLogger:
    """
    Wrapper class designed to work as a method
    decorator, making logging of output and caught
    errors easier.

    Begin with creating the logging instance of
    choice, configuring it the way you want, then
    pass it to the logger.set_logger method.

    @pyttman.logger
    def myfunc(self, *args, **kwargs):
        ...

    log (method):
        If you want to manually log custom messages in your code,
        you can call this method. See method docstring / help
        for parameters and how to use it.
    """

    LOG_INSTANCE = None

    def __call__(self, func):
        """
        Wrapper method for providing logging functionality.
        Use @logger to implement this method where logging
        of methods are desired.
        """
        def inner(*args, log_level="debug", log_exception=True, **kwargs):
            """
            Inner method, executing the func parameter function,
            as well as executing the logger.
            :returns:
                Output from executed function in parameter func
            """
            PyttmanLogger._verify_config_complete()
            try:
                results = func(*args, **kwargs)
                message = f"Return value from '{func.__name__}': '{results}'"
                self.log(message=message, level=log_level)
                return results
            except Exception as e:
                if log_exception:
                    message = (f"Exception occurred in {func.__name__}. "
                               f"Traceback: {traceback.format_exc()} {e}")
                    self.log(message=message, level="error")
                raise e
        return inner

    @staticmethod
    def _verify_config_complete():
        if pyttman.logger.LOG_INSTANCE is None:
            raise RuntimeError('Internal Pyttman Error: '
                               'No Logger instance set.\r\n')

    def loggedmethod(self, func: callable):
        """
        Backward compatibility only; use @logger
        """
        def inner(*args, **kwargs):
            return self.__call__(func)(*args, **kwargs)
        return inner

    def log(self, message: str, level="debug") -> None:
        """
        Allow for manual logging during runtime.
        """
        PyttmanLogger._verify_config_complete()
        log_levels = {"info": self.LOG_INSTANCE.info,
                      "debug": self.LOG_INSTANCE.debug,
                      "error": self.LOG_INSTANCE.error}
        try:
            log_levels[level](message)
        except KeyError:
            log_levels["debug"](message)



import inspect
import traceback
import uuid
import warnings
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any

from pyttman_base_plugin import PyttmanPlugin

import pyttman
from pyttman.core.containers import MessageMixin, Reply
from pyttman.core.decorators import LifecycleHookRepository
from pyttman.core.exceptions import PyttmanPluginException
from pyttman.core.mixins import PrettyReprMixin


def depr_raise(message: str, version: str) -> None:
    """
    Raise DeprecationWarning with a message and version tag for users.
    :param message: Deprecation message to display to users
    :param version: Pyttman version in which deprecation was declared
    :raise DeprecationWarning
    """
    out = f"{message} - This was deprecated in version {version}."
    raise DeprecationWarning(out)


def depr_graceful(message: str, version: str):
    """
    Uses warnings.warn with a message and version tag for users.
    :param message: Deprecation message to display to users
    :param version: Pyttman version in which deprecation was declared
    """
    out = f"{message} - This was deprecated in version {version}."
    warnings.warn(out, DeprecationWarning)


class Settings:
    """
    Dataclass holding settings configured in the settings.py
    settings module. Modules are not picklable in Python,
    thus this class holds the user-level variables and
    objects created in the module instead of the flooding
    of using the entire module as reference in 'pyttman.settings'
    throughout Pyttman apps.

    The Settings class automatically omits any instance
    in **kwargs being of instance <module> since modules
    aren't picklable. It also omits functions as callables
    aren't valid settings.
    """

    def __init__(self, dictionary=None, **kwargs):
        if dictionary is None:
            dictionary = {}
        self.__dict__.update(dictionary)
        self.APPEND_LOG_FILES: bool = True
        self.ROUTER: dict | None = None
        self.ABILITIES: list | None = None
        self.FATAL_EXCEPTION_AUTO_REPLY: list | None = None
        self.CLIENT: dict | None = None
        self.APP_BASE_DIR: str | None = None
        self.LOG_FILE_DIR: str | None = None
        self.APP_NAME: str | None = None
        self.LOG_FORMAT: str | None = None
        self.LOG_TO_STDOUT: bool = False
        self.STATIC_FILES_DIR: Path | None = None
        self.PLUGINS: list | None = None

        [setattr(self, k, v) for k, v in kwargs.items()
         if not inspect.ismodule(v)
         and not inspect.isfunction(v)]

    def __repr__(self):
        _attrs = {name: value for name, value in self.__dict__.items()}
        return f"Settings({_attrs})"


def _generate_name(name):
    """
    Generates a user-friendly name out of
    Command or Ability class names, by
    inserting spaces in camel cased names
    as well as truncating 'Command' and 'Ability'
    in the names.
    :param name: string, name of a class.
                 hint: Command or Ability subclass
    :return: str, 'SetTimeCommand' -> 'Set Time'
    """
    new_name = ""
    for i in ("Ability", "feature", "Command", "command"):
        name = name.replace(i, "")

    for i, c in enumerate(name):
        if i > 0 and c.isupper():
            new_name += " "
        new_name += c
    return new_name


def _generate_error_entry(message: MessageMixin, exc: BaseException) -> Reply:
    """
    Creates a log entry for critical errors with a UUID bound
    to the log file entry, explaining the error. For the front
    end clients, a Reply object is returned to provide for
    end users who otherwise would experience a chatbot who
    didn't reply at all.
    :param message: MessageMixin
    :param exc: Exception
    :return: Reply
    """
    error_id = uuid.uuid4()
    traceback.print_exc()
    warnings.warn(f"{datetime.now()} - A critical error occurred in the "
                  f"application logic. Error id: {error_id}")
    error_message = (f"CRITICAL ERROR: ERROR ID={error_id} - "
                     f"The error was caught while processing message: "
                     f"'{message}'. Error message: '{exc}'")
    try:
        pyttman.logger.log(level="error", message=error_message)
    except Exception:
        print(error_message)

    try:
        auto_reply = pyttman.settings.MIDDLEWARE['FATAL_EXCEPTION_AUTO_REPLY']
    except Exception:
        auto_reply = "An internal error occurred in the application."
    return Reply(f"{auto_reply} ({error_id})")


@dataclass
class PyttmanApp(PrettyReprMixin):
    """
    The highest point of abstraction for a Pyttman application.
    This class holds the Settings, the Abilities and lifecycle hooks
    for the application, including the Client class used to interface with
    the platform of choice.
    This singleton instance is available through 'from pyttman import app'.
    """
    __repr_fields__ = ("name", "client", "hooks")

    @dataclass
    class LoadedPluginContainer:
        """
        Access point for loaded plugins during the app runtime, e.g:
        `from pyttman import app`
        `my_plugin = app.loaded_plugins.SomePluginName`
        """

        def ingest(self, plugin):
            setattr(self, plugin.__class__.__name__, plugin)

    client: Any
    name: str | None = field(default=None)
    settings: Settings | None = field(default=None)
    hooks: LifecycleHookRepository = field(
        default_factory=lambda: LifecycleHookRepository())
    _abilities: set = field(default_factory=set)
    plugins: list[PyttmanPlugin] = field(default_factory=list)
    loaded_plugins: LoadedPluginContainer = field(default_factory=LoadedPluginContainer)

    def __post_init__(self):
        for plugin in self.plugins:
            plugin.app = self
            plugin.on_app_start()
            self.loaded_plugins.ingest(plugin)

    def start(self):
        """
        Start a Pyttman application.
        """
        # noinspection PyBroadException
        # Execute plugin hooks
        self.execute_plugins_before_app_start()
        try:
            self.client.run_client()
        except Exception:
            warnings.warn(traceback.format_exc())
        self.execute_plugins_after_app_stop()

    @property
    def abilities(self):
        return self._abilities

    @abilities.setter
    def abilities(self, abilities):
        for ability in abilities:
            setattr(self, ability.__class__.__name__, ability)
            self._abilities.add(ability)

    def raise_if_none(self, message: MessageMixin, plugin: PyttmanPlugin, method: callable):
        """
        Raise a ValueError if the message is None.
        """
        if message is None:
            raise PyttmanPluginException(f"Plugin malfunction: {plugin} returned None "
                                         f"instead of a MessageMixin object when executing"
                                         f"method '{method}'.")

    def execute_plugins_before_router(self, message: MessageMixin):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PyttmanPlugin.PluginInterceptPoint.before_router):
                message = plugin.before_router(message)
                self.raise_if_none(message, plugin, plugin.before_router)
        return message

    def execute_plugins_before_entity_extraction(self, message: MessageMixin):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PyttmanPlugin.PluginInterceptPoint.before_entity_extraction):
                message = plugin.before_entity_extraction(message)
                self.raise_if_none(message, plugin, plugin.before_entity_extraction)
        return message

    def execute_plugins_after_entity_extraction(self, message: MessageMixin):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PyttmanPlugin.PluginInterceptPoint.after_entity_extraction):
                message = plugin.after_entity_extraction(message)
                self.raise_if_none(message, plugin, plugin.after_entity_extraction)
        return message

    def execute_plugins_before_intent(self, message: MessageMixin):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PyttmanPlugin.PluginInterceptPoint.before_intent):
                message = plugin.before_intent(message)
                self.raise_if_none(message, plugin, plugin.before_intent)
        return message

    def execute_plugins_after_intent(self, reply: Reply):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PyttmanPlugin.PluginInterceptPoint.after_intent):
                reply = plugin.after_intent(reply)
                self.raise_if_none(reply, plugin, plugin.after_intent)
        return reply

    def execute_plugins_before_app_start(self):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PPyttmanPlugin.PluginInterceptPoint.before_app_start):
                plugin.before_app_start(self)

    def execute_plugins_after_app_stop(self):
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PPyttmanPlugin.PluginInterceptPoint.after_app_stops):
                plugin.after_app_stops(self)

    def execute_plugins_no_intent_match(self, message: MessageMixin):
        reply = None
        for plugin in self.plugins:
            if plugin.allowed_to_intercept_at(PPyttmanPlugin.PluginInterceptPoint.no_intent_match):
                reply = plugin.no_intent_match(message)
                self.raise_if_none(message, plugin, plugin.no_intent_match)
        return reply



import re
from copy import copy
from dataclasses import dataclass
from datetime import datetime
from queue import Queue
from typing import List, Iterable

from pyttman.core.mixins import PrettyReprMixin


class MessageMixin(PrettyReprMixin):
    """
    Pyttman MessageMixin, to extend the functionality
    of existing Message classes provided by 3rd party
    libraries and APIs, to also accommodate for the
    internal requirements of the Message object
    which is expected to fulfill a certain contract
    of attributes and methods for parsing messages.

    The MessageMixin class can be included in multiple
    inheritance when a Message-like class is developed
    for supporting a 3rd party library / API.
    """
    __repr_fields__ = ("author", "user", "created", "entities")

    @dataclass
    class Author:
        id: int = None

    def __init__(self, content=None, **kwargs):
        self.author = self.Author()
        self.user = None
        self.created = datetime.now()
        self.client = None
        self.content = content
        self.entities = {}

        try:
            self.content_with_format = str(content).splitlines(keepends=True)
        except ValueError:
            self.content_with_format = None

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.content == other.content
        return False

    def __getitem__(self, index: int) -> str:
        return self.content[index]

    def segmented(self, batch_size: int):
        """
        Returns a generator that yields the content
        of the message in segments of 'batch_size'.
        The generator will yield the content closest to
        a complete word, and not split words in half,
        which is why the generator may yield less than
        'batch_size' elements.
        """
        buf = []
        content = copy(self.content)
        while content:
            if len(" ".join(buf)) + len(content[0]) < batch_size:
                buf.append(content.pop(0))
            else:
                yield " ".join(buf)
                buf = []
        yield " ".join(buf)

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, val):
        if val is None:
            self._content = ["None"]
        elif isinstance(val, str):
            self._content = val.split()
        elif isinstance(val, list) or isinstance(val, tuple):
            self._content = [str(i) for i in val]
        elif isinstance(val, dict):
            self._content = str(val).split()
        else:
            try:
                self._content = repr(val).split()
            except Exception:
                raise TypeError(f"content cannot be type {type(val)} "
                                f"as it is could not be typecast to "
                                f"str.")

    def sanitized_content(self, preserve_case=False) -> List[str]:
        """
        Return a sanitized version of the .content property.
        This means that the contents in the message
        are stripped of all symbols while digits are still kept.
        Case is preserved if preserve_case is True.
        :return: list
        """
        out = []
        for i in self.content:
            sanitized = re.sub(r"[^\w\s]", "", i)
            if preserve_case:
                out.append(sanitized)
            else:
                out.append(sanitized.lower())
        return out

    def lowered_content(self) -> List[str]:
        """
        Returns the content of the message case lowered.
        :return: list, str
        """
        return [i.lower() for i in self.content]

    def as_str(self, sanitized: bool = False) -> str:
        """
        Return the 'content' field as joined string
        :param sanitized: Return the content as sanitized_content or not
        :return: str
        """
        if sanitized:
            return " ".join(self.sanitized_content())
        elif self.content_with_format is not None:
            return str().join(self.content_with_format)
        return " ".join(self.content)

    def as_list(self, sanitized: bool = False) -> List:
        """
        Return the 'content' field as list
        :return: list
        """
        if sanitized:
            content = self.sanitized_content()
        else:
            content = self.content

        if isinstance(content, list):
            return content
        elif isinstance(content, str):
            return content.split()

    def remove(self, item):
        """
        Removes element from self.content.
        :return: None
        """
        self.content.remove(item)


class Message(MessageMixin):
    """
    Standard implementation of the MessageMixin
    class without extending any functionality.
    """
    pass


class Reply(MessageMixin):
    """
    The Reply object is expected to be  returned
    from all Intent subclasses.
    """
    pass


class ReplyStream(Queue):
    """
    The ReplyStream class can be used instead of
    the Reply class, whenever a collection of
    Reply objects are to be returned to the
    user. 
    """

    def __init__(self, collection: Iterable = None):
        super().__init__()
        if collection is not None:
            if isinstance(collection, str):
                self.put(collection)
            else:
                try:
                    iter(collection)
                except Exception:
                    raise AttributeError("'collection' must be iterable")
                else:
                    [self.put(i) for i in collection]

    def get(self, block=True, timeout=None):
        """
        Remove and return an item from the ReplyStream.
        """
        element = self._get()
        return Reply(element) if not isinstance(element, Reply) else element



import re
import string
import typing
from itertools import zip_longest
from typing import Type, Dict, Union

from ordered_set import OrderedSet

from pyttman.core.containers import MessageMixin
from pyttman.core.entity_parsing.entity import Entity
from pyttman.core.entity_parsing.identifiers import Identifier
from pyttman.core.exceptions import InvalidPyttmanObjectException
from pyttman.core.mixins import PrettyReprMixin


class EntityFieldValueParser(PrettyReprMixin):
    """
    This class is used by EntityField classes primarily,
    as the inner-working engine for identifying and finding
    values which match the pattern provided in the declarative
    EntityParser Api component: 'EntityField'.
    """
    __repr_fields__ = ("identifier", "exclude", "prefixes", "suffixes")
    ignore_chars = True
    chars_to_ignore = ".,;:!?-"

    def __init__(self,
                 prefixes: tuple | typing.Callable = None,
                 suffixes: tuple | typing.Callable = None,
                 valid_strings: tuple | typing.Callable = None,
                 default:  typing.Any | typing.Callable = None,
                 span: int | typing.Callable = 0,
                 identifier: Type[Identifier] | None = None,
                 exclude: typing.Iterable[str] = None,
                 as_list: bool = False,
                 **kwargs):

        if hasattr(self, "value"):
            raise AttributeError("The field 'value' is reserved for internal "
                                 "use. Please choose a different name for "
                                 "the field.")

        self.truncates_message_in_parsing = True
        self.value = None
        self.case_preserved_cache = set()
        self.prefixes = prefixes or tuple()
        self.suffixes = suffixes or tuple()
        self.exclude = exclude or tuple()
        self.valid_strings = valid_strings or tuple()
        self.default = default if default is not None else None
        self.identifier = identifier
        self.span = span
        self.as_list = as_list
        self._properties_for_evaluation = {
            "prefixes": self.prefixes,
            "suffixes": self.suffixes,
            "span": self.span,
            "default": self.default,
            "valid_strings": self.valid_strings
        }

    def _prepare_params(self):
        """
        Prepares the arguments passed in the constructor.
        This happens after init in order to allow developers to provide
        callables as arguments, which should not be evaluated / executed
        at __init__ time due to the risk of DB connections not being
        established at that time.
        :return:
        """
        for name, value in self._properties_for_evaluation.items():
            if callable(value):
                try:
                    setattr(self, name, value())
                except Exception as e:
                    raise InvalidPyttmanObjectException(
                        "An error occurred during preparation of field "
                        f"'{name}' in '{self}'. If the callable takes "
                        f"arguments without provided defaults, consider "
                        f"using a partial.")

        # Validate that the object was constructed properly
        if not isinstance(self.prefixes, tuple) or \
                not isinstance(self.suffixes, tuple):
            raise AttributeError(f"'{self}' "
                                 f"is incorrectly configured: "
                                 f"'prefixes' and 'suffixes' "
                                 f"must be tuples.\nDo you have "
                                 f"a tuple with only one item? "
                                 f"Don't forget the trailing comma, "
                                 f"example: '(1,)' instead of '(1)'.")

        if self.valid_strings is not None and not isinstance(self.valid_strings, typing.Sequence):
            raise AttributeError("'valid_strings' must be a collection of "
                                 f"strings, got: '{self.valid_strings}'")
        else:
            self.valid_strings = tuple([i.casefold() for i in self.valid_strings])

    def reset(self) -> None:
        """
        Resets the parser, defaulting it's value to None.
        :return: None
        """
        self.value = None

    def parse_message(self,
                      message: MessageMixin,
                      original_message_content: tuple[str],
                      memoization: dict = None) -> None:
        """
        Walk the message and parse it for values.
        If the identified value exists in memoization,
        traverse on until value is returned or the
        message is exhausted.

        :param message: A Message object to pass to each EntityField,
        for parsing entities.
        :param original_message_content: The original untouched contents of
        the message as received from the client. This is used for a source
        of truth, since the content in `message` is mutated with each entity
        field parsing it.
        :param memoization: Dictionary with previously identified entities
        """
        self._prepare_params()
        if self.valid_strings:
            output = []
            word_index = 0

            message_lowered = message.lowered_content()
            if self.ignore_chars:
                # Strip away special chars
                message_lowered = [
                    re.sub(rf"[{self.chars_to_ignore}]", "", word)
                    for word in message_lowered
                ]

            common_occurrences = tuple(
                OrderedSet(message_lowered).intersection(self.valid_strings))

            for i, word in enumerate(common_occurrences):
                if i > self.span and not self.as_list:
                    break
                word_index = message_lowered.index(word)
                output.append(message.content[word_index])

            if len(output) > 1:
                self.value = Entity(output, index_in_message=word_index)
            elif len(output) == 1:
                self.value = Entity(output.pop())
            else:
                self.value = Entity(self.default, is_fallback_default=True)

            self._validate_prefixes_suffixes(message)

            if self.value:
                if self.ignore_chars:
                    if isinstance(self.value.value, list):
                        for i, elem in enumerate(self.value.value):
                            elem = re.sub(rf"[{self.chars_to_ignore}]", "", elem)
                            self.value.value[i] = elem
                    elif isinstance(self.value.value, str):
                        self.value.value = re.sub(rf"[{self.chars_to_ignore}]", "", self.value.value)
                entity = self.value
                if entity.value == self.default:
                    return
                if isinstance(entity.value, list):
                    words_to_remove_from_message = entity.value
                    entity.index_in_message += len(entity.value)
                elif isinstance(entity.value, str):
                    words_to_remove_from_message = [entity.value]

                self._remove_words_from_message_unless_default(
                    message,
                    *words_to_remove_from_message)
            return

        if self.truncates_message_in_parsing is False:
            return

        for i, _ in enumerate(message.content):
            parsed_entity: Entity = self._identify_value(message,
                                                         start_index=i)
            # An entity has been identified, and it's unique.
            if parsed_entity is not None and memoization.get(
                    parsed_entity.index_in_message) is None:
                self.value = parsed_entity
                break
            else:
                self.reset()

    def _remove_words_from_message_unless_default(
            self,
            message,
            *words_to_remove_from_message) -> None:
        """
        Removes elements from message.content, if they're present
        in the message, unless a word or other object happens to
        equal the 'self.default' property on the instance.
        """
        for word in words_to_remove_from_message:
            if word == self.default:
                continue
            try:
                message.content.remove(word)
            except ValueError:
                continue

    def _validate_message_with_affixes(self,
                                       affixes: tuple[str],
                                       message: MessageMixin,
                                       comparator: callable):
        entity = self.value
        if not (common_strings := set(affixes).intersection(message.content)):
            entity.value = self.default
            return
        for string in common_strings:
            if not comparator(message.content.index(string), entity.index_in_message):
                entity.value = self.default

    def _validate_prefixes_suffixes(self, message: MessageMixin):
        """
        Check 'prefixes' and 'suffixes' for Entity values to make sure
        that they comply
        :return:
        """
        if self.prefixes:
            self._validate_message_with_affixes(
                self.prefixes,
                message,
                lambda affix_index, value_index: affix_index < value_index)
        if self.suffixes:
            self._validate_message_with_affixes(
                self.suffixes,
                message,
                lambda affix_index, value_index: affix_index > value_index)

    def _identify_value(self, message: MessageMixin,
                        start_index: int = 0) -> Union[None, Entity]:
        """
        Parses the message for values to identify.

        Since the parser can be configured in many ways;
            * With suffixes only
            * With prefixes only
            * With prefixes and suffixes
            * With or without an Identifier class in all of the above

        ... these conditions need to be evaluated for each scenario.

        Pre- and Suffix tuples can contain strings or Parsers, or a
        combination. When EntityFieldValueParser instances are used in these tuples,
        their last occurring string, separated by spaces, is chosen
        as the ultimate prefix. This is so due to the fact that Parsers
        may span across multiple elements in the string collection
        and take ownership of more than one string in the sequence.

        If an Identifier is provided, it will try and find a value
        which complies with the pre- and suffixes as provided.
        If the Identifier is alone, it will return the first value
        encountered in the traversing of the message.

        :param message: Message object to parse
        :param start_index: Index pointer, where parsing
                            is started from in the message
        :return: Entity or None, depending on if parsing is successful.
        """
        prefix_entity = None
        suffix_entity = None
        prefixes = []
        suffixes = []
        prefix_indices = []
        suffix_incides = []
        last_prefix_index, earliest_suffix_index = 0, 0
        parsed_entity: Union[Entity, None] = None
        lowered_msg_content = message.lowered_content()

        # First - traverse over the pre- and suffixes and
        # collect them in separate lists
        for i_prefix, i_suffix in zip_longest(self.prefixes,
                                              self.suffixes,
                                              fillvalue=None):

            for rule, rule_collection in {i_prefix: prefixes,
                                          i_suffix: suffixes}.items():
                if rule is not None:
                    if isinstance(rule, EntityFieldValueParser) and rule.value is not None:
                        entity: Entity = rule.value
                        rule_collection.append(entity.value
                                               .split().pop().lower().strip())
                    elif isinstance(rule, str):
                        rule_collection.append(rule)

        for prefix, suffix in zip_longest(prefixes, suffixes, fillvalue=None):
            try:
                # Save the index of this prefix in the message
                if prefix is not None:
                    prefix_indices.append(lowered_msg_content.index(prefix))
                if suffix is not None:
                    suffix_incides.append(lowered_msg_content.index(suffix))
            except ValueError:
                # The prefix was not in the message
                continue

        # Let's extract the last occurring prefix,
        # and the earliest suffix of the ones present
        if len(prefixes):
            if not len(prefix_indices):
                return None
            last_prefix_index = max(prefix_indices)
            try:
                index_for_value = last_prefix_index + 1
                value_at_index = message.content[index_for_value]
                prefix_entity = Entity(value=value_at_index,
                                       index_in_message=index_for_value)
            except IndexError:
                prefix_entity = None
            start_index = last_prefix_index + 1

        if len(suffixes):
            if not len(suffix_incides):
                return None

            earliest_suffix_index = min(suffix_incides)
            try:
                index_for_value = earliest_suffix_index - 1
                value_at_index = message.content[index_for_value]
                suffix_entity = Entity(value=value_at_index,
                                       index_in_message=index_for_value)
            except IndexError:
                suffix_entity = None

        # If an Identifier was used - let it parse the message, but make
        # sure it complies with Pre- and/or suffix values, if configured
        if self.identifier is not None:
            identifier_object = self.identifier(start_index=start_index)
            identifier_entity = identifier_object.try_identify_entity(message)

            if identifier_entity is not None:
                allowed_scenarios = {
                    bool(not self.prefixes and not self.suffixes),

                    bool(self.prefixes and identifier_entity
                         .index_in_message >= (last_prefix_index + 1)),

                    bool(self.suffixes and identifier_entity.
                         index_in_message <= (earliest_suffix_index - 1)),

                    bool(self.prefixes and self.suffixes and
                         last_prefix_index < identifier_entity.
                         index_in_message < earliest_suffix_index)
                }

                if any(allowed_scenarios):
                    parsed_entity = identifier_entity
        else:
            if self.prefixes and not self.suffixes:
                parsed_entity = prefix_entity
            elif self.suffixes and not self.prefixes:
                parsed_entity = suffix_entity
            elif self.prefixes and self.suffixes:
                try:
                    begin = prefix_entity.index_in_message
                    end = suffix_entity.index_in_message
                    parsed_entity = Entity(
                        value=message.content[begin:end].pop(),
                        index_in_message=begin + 1)
                except IndexError:
                    parsed_entity = None

        # The ValueParser has no prefixes, suffixes or identifier.
        # The entity is the first string in the message.
        if not prefixes and not suffixes and not self.identifier:
            parsed_entity = Entity(value=message.content[0],
                                   index_in_message=0)

        # If the span property is set to greater than 0, walk further in
        # message.content and also include elements as far as the span
        # property designates.
        # If an identifier is used, it also has to approve of the string
        # for each span iteration as the walk in the message progresses.
        # If an Identifier is does not comply with a string, the walk is
        # cancelled.
        if parsed_entity is None:
            return parsed_entity

        while parsed_entity.value.lower() in [i.lower() for i in self.exclude]:
            parsed_entity.index_in_message += 1
            # Traverse the message for as long as the current found
            # entity is in the 'exclude' tuple. If the end of message
            # is reached, quietly break the loop.
            try:
                parsed_entity.value = message.content[
                    parsed_entity.index_in_message]
            except IndexError:
                return None

        # Now, add words for as long as `span` allows us to iterate.
        for i in range(1, self.span):
            parsed_entity.index_in_message += 1
            try:
                if self.identifier:
                    identifier_object: Identifier = self.identifier(
                        start_index=parsed_entity.index_in_message)
                    # Identifier did not find
                    span_entity = identifier_object.try_identify_entity(
                        message)
                    if span_entity is None or span_entity.index_in_message != parsed_entity.index_in_message:
                        break
                    span_value = span_entity.value
                else:
                    span_value = message.content[parsed_entity.index_in_message]

            # There are not enough elements in message.content to walk
            # as far as the span property requests - abort.
            except IndexError:
                break
            else:
                if span_value not in self.exclude:
                    parsed_entity.value += f" {span_value}"
        return parsed_entity


def parse_entities(message: MessageMixin,
                   entity_fields: dict[str, EntityFieldValueParser],
                   original_message_content: tuple[str],
                   exclude: tuple = None) -> dict[str, str]:
    """
    Traverse over all fields which are EntityFieldValueParser subclasses.
    Have them identify their values according to their
    constraints and conditions, and store them in a
    dictionary, returned at the end of parsing.
    :param original_message_content: The original untouched contents of
        the message as received from the client. This is used for a source
        of truth, since the content in `message` is mutated with each entity
        field parsing it.
    :param exclude: Optional tuple of strings to ignore in parsing.
    :param entity_fields: Dictionary with `name: EntityField` mapped
    :param message: MessageMixin subclass object to be parsed.
    :return: Dictionary with the name of the entity against its parsed value.
    """
    output = {}

    if exclude is None:
        exclude = []
    elif isinstance(exclude, tuple):
        exclude = list(exclude)

    # The memoization dict is provided each EntityFieldValueParser instance
    # in order for them to avoid catching a string, previously
    # caught by a predecessor in iterations.
    parsers_memoization: Dict[int, Entity] = {}
    parser_joined_suffixes_and_prefixes: typing.Set[str] = set()

    for field_name, entity_field_instance in entity_fields.items():

        exclude.extend(entity_field_instance.exclude)
        # Collect all parser pre- and suffixes
        parser_joined_suffixes_and_prefixes.update(
            entity_field_instance.prefixes + entity_field_instance.suffixes)

        # Share the 'exclude' tuple assigned by the developer in the
        # application code to each EntityFieldValueParser instance
        entity_field_instance.exclude = exclude
        entity_field_instance.parse_message(
            message,
            memoization=parsers_memoization,
            original_message_content=original_message_content)

        # See what the parser found - Entity or None.
        # Ignore entities in self.exclude.
        parsed_entity: Union[Entity, None] = entity_field_instance.value

        if parsed_entity is None or parsed_entity.value in exclude:
            # Use the developer declared fallback value (None, by default)
            output[field_name] = Entity(entity_field_instance.default,
                                        is_fallback_default=True)
        else:
            parsed_entity.as_list = entity_field_instance.as_list
            output[field_name] = parsed_entity

            # Store the entity for memoization to
            # prohibit multiple occurrences
            if entity_field_instance.truncates_message_in_parsing:
                parsers_memoization[
                    parsed_entity.index_in_message] = parsed_entity

    """
    Walk the message backwards and truncate entities which 
    contain elements from entities occurring later in the 
    message. 

    All elements in pre- and suffixes are also truncated 
    from all entities as they are delimiters, and should 
    not be present in the entity value.
    """
    duplicate_cache: typing.Set[str] = set(
        parser_joined_suffixes_and_prefixes)

    for field_name, entity in reversed(output.items()):
        if entity.is_boolean():
            continue

        entity_field = entity_fields.get(field_name)
        duplicate_cache.update(entity_field.case_preserved_cache)
        value_for_type_conversion = entity_field.default

        # Assess only Parsers which have successfully parsed entities.
        if entity.value is not None and entity.is_fallback_default is \
                False:
            # Value is a string - split the entity value by space,
            # so we can work with it
            try:
                split_value = entity.value.split()
            except AttributeError:
                split_value = entity.value

            # Truncate prefixes, suffixes and cached strings
            # from the entity
            if split_value != list(duplicate_cache):
                split_value = OrderedSet(split_value) \
                    .difference(duplicate_cache)

            duplicate_cache.update(split_value)
            duplicate_cache.update(set([i.casefold()
                                        for i in split_value]))
            if not entity.as_list:
                value_for_type_conversion = str(" ").join(split_value)
            else:
                value_for_type_conversion = entity.value

        # New in 1.1.9 - If this is an EntityField class, convert
        # the value in the Entity with it.
        try:
            if entity.as_list:
                if isinstance(entity.value, list):
                    entity.value = [entity_field.convert_value(i) for i in entity.value]
                else:
                    entity.value = [entity_field.convert_value(value_for_type_conversion)]
            else:
                entity.value = entity_field.convert_value(value_for_type_conversion)
        except AttributeError:
            entity.value = value_for_type_conversion

        output[field_name] = entity
    case_converted_output = [i.lower() for i in output]

    for word in exclude:
        if case_converted_output.count(word):
            output.remove(output[case_converted_output.index(word)])

    return output


#   Backwards compatibility, Pyttman<=1.1.9.1
if __name__ != "__main__":
    ValueParser = EntityFieldValueParser



import asyncio
import functools
import inspect
import warnings
from datetime import datetime
from queue import Queue
from typing import Dict, Generator, Any, Callable, Tuple
from multidict import MultiDict


from pyttman.tools.scheduling.components import Job, TimeTrigger


# noinspection PyPep8Naming
class schedule:
    """
    Namespacing static class to schedule
    function calls, using TimeTrigger and Job
    objects.

    Works as a decorator using the
    'method' method as @Schedule.method,
    or by calling Schedule.run() with
    provided args.
    """
    id_job_map: Dict[int, Job] = {}
    name_job_map: MultiDict[str, Job] = MultiDict()
    outputs: Queue[Job] = Queue()

    @staticmethod
    def method(func, at: str = None, every: str = None,
               delay=None, exactly_at: datetime = None,
               recipient: Callable = None, start_now=True,
               async_loop=None, **kwargs) -> Job:
        """
        Registers a new schedule Job. Provide strings
        to define when to execute the job, if it is
        recurring and whether to pass the output to
        another callable as 'recipient'.

        example:
            schedule.method(every="day", at="10:30",
                            func=some_func,
                            some_arg="hello world")

        With defined recipient which will receive the
        output from the job as the first arg:

        example:
            schedule.method(every="day", at="10:30",
                            func=some_func,
                            some_arg="hello world",
                            recipient=print)

        Users can provide an async loop to schedule an async task
        in, but it is not mandatory - this may cause problems
        however in context managed async tasks which requires
        access to the loop to retrieve tasks themselves.
        TL;DR - if problems occur when 'async_loop' argument
        is omitted - provide the loop for where to schedule
        the recipient coroutine.

        @param func: Callable to be scheduled - unbound or bound, sync or async
        @param at: str, Timestamp for execution, "HH:MM" ("HH:MM:SS optionally)
        @param every: str Options: monday -> friday, day, hour, minute, second
        @param delay: str, Delay execution, "HH:MM:SS" ("HH:MM:SS optionally)
        @param exactly_at: datetime object, Exact point in time for execution
        @param recipient: callable to which the job output will be passed to
        @param start_now: Start the job now, default: True. If not,
                          see schedule.unstarted
        @param async_loop: optional loop to schedule job and/or recipient in
        @param kwargs: kwargs, passed to the function in 'func' when creating
                       the partial function which is then passed to the Job
        @return: Job instance

        """
        async_warn = "\nThe callable '{0}' is asynchronous but the " \
                     "'async_loop' argument was omitted, and no async " \
                     "loop is running.\nThis will cause the job to be " \
                     "executed through 'asyncio.run().'\n"
        return_self = False

        # Create a TimeTrigger for the Job
        trigger = TimeTrigger(every=every, at=at,
                              delay=delay,
                              exactly_at=exactly_at)

        if not (recipient := recipient):
            recipient = schedule.schedule_default_catcher
            return_self = True

        func_is_async = inspect.iscoroutinefunction(func)
        recipient_is_async = inspect.iscoroutinefunction(recipient)
        if func_is_async or recipient_is_async:
            if async_loop is None:
                try:
                    async_loop = asyncio.get_running_loop()
                # No loop is running, and none was provided.
                except RuntimeError:
                    warnings.warn(async_warn.format(func))

        if inspect.iscoroutinefunction(recipient) and async_loop is None:
            warnings.warn(async_warn.format(recipient))

        job = Job(func=functools.partial(func, **kwargs),
                  is_async=func_is_async,
                  trigger=trigger,
                  recipient=recipient,
                  func_name=repr(func),
                  async_loop=async_loop,
                  return_self=return_self)

        # Map job in schedule and start it
        schedule.name_job_map.add(job.func_name, job)

        # Start the job, or add it to unstarted for later starts
        if start_now:
            job.start()
            schedule.id_job_map[job.native_id] = job
        return job

    @staticmethod
    def schedule_default_catcher(job: Job) -> None:
        """
        This is the default method for return values of
        schedule jobs, if no recipient is specified
        by the creator of a schedule job.

        All Jobs without a designated recipient
        callable will have their return value end
        up in this method which maps their return
        value with themselves as source, in a dict object
        and stored in the public accessible queue "outputs".

        :param job: Job instance which has been executed
                    at least once
        """
        schedule.outputs.put(job)

    @staticmethod
    def get_jobs(job_name: str = None, job_id: int = None) -> \
            Generator[Job, Any, None]:
        """
        Returns Job instance(s) that matches provided
        args. Since many jobs can share the same name,
        multidict is used when mapping against its name.
        Their native_id is unique however, thus a regular
        dict is used for mapping jobs against their native_id.


        :param job_name: str, name of the job to return (optional)
        :param job_id: int, id of the job to return (optional
        :rtype: Job instance
        :raises: ValueError, if both name and id are None
        """
        if job_name is None and id is None:
            raise ValueError("either job_name or job_id must be specified")

        if job_name:
            try:
                for job in schedule.name_job_map.getall(job_name):
                    yield job
            except KeyError:
                yield
        elif job_id:
            try:
                yield schedule.id_job_map[int(job_id)]
            except KeyError:
                yield

    @staticmethod
    def kill_job_gracefully(job_name=None, job_id=None):
        """
        Kill jobs gracefully.
        :param job_name: str, name of the job
        :param job_id: int, id of the job
        """
        for job in schedule.get_jobs(job_name, job_id):
            job.kill_gracefully()

    @staticmethod
    def get_all_jobs() -> Generator[Job, Any, None]:
        """
        Returns all jobs, in name_job_map and id_job_map
        combined.
        :rtype: Job
        """
        for i in schedule.name_job_map.values():
            yield i

    @staticmethod
    def has_outputs() -> bool:
        """
        Returns whether there are outputs in
        the output queue 'schedule.outputs'
        """
        return bool(schedule.outputs.qsize())

    @staticmethod
    def get_latest_output():
        return schedule.outputs.get()

    @staticmethod
    def get_unstarted_jobs(name: str = None) -> Tuple[Job]:
        """
        Returns collection of all jobs that are not
        running at the moment.
        @param name: Job name (may return multiple)
        @return: Tuple with Jobs
        """
        if name:
            return tuple([i for i in schedule.get_jobs(name)
                          if i.running is False])
        return tuple([i for i in schedule.get_all_jobs()
                      if i.running is False])

    @staticmethod
    def start_job(name: str) -> None:
        """
        Start the scheduling of a given method,
        if a Job with the name is found
        @param name: Name of the job to start (may start multiple)
        @return: None
        """
        for job in schedule.get_jobs(job_name=name):
            job.start()
            schedule.id_job_map[job.native_id] = job
